searchNodes=[{"doc":"Broadway is a concurrent, multi-stage tool for building data ingestion and data processing pipelines. It allows developers to consume data efficiently from different sources, such as Amazon SQS, Apache Kafka, Google Cloud PubSub, RabbitMQ and others. Built-in features Back-pressure - by relying on GenStage, we only get the amount of events necessary from upstream sources, never flooding the pipeline. Automatic acknowledgements - Broadway automatically acknowledges messages at the end of the pipeline or in case of errors. Batching - Broadway provides built-in batching, allowing you to group messages either by size and/or by time. This is important in systems such as Amazon SQS, where batching is the most efficient way to consume messages, both in terms of time and cost. Fault tolerance with minimal data loss - Broadway pipelines are carefully designed to minimize data loss. Producers are isolated from the rest of the pipeline and automatically resubscribed to in case of failures. On the other hand, user callbacks are stateless, allowing us to handle any errors locally. Finally, in face of any unforeseen bug, we restart only downstream components, avoiding data loss. Graceful shutdown - Broadway integrates with the VM to provide graceful shutdown. By starting Broadway as part of your supervision tree, it will guarantee all events are flushed once the VM shuts down. Built-in testing - Broadway ships with a built-in test API, making it easy to push test messages through the pipeline and making sure the event was properly processed. Custom failure handling - Broadway provides a handle_failed/2 callback where developers can outline custom code to handle errors. For example, if they want to move messages to another queue for further processing. Dynamic batching - Broadway allows developers to batch messages based on custom criteria. For example, if your pipeline needs to build batches based on the user_id, email address, etc, it can be done by calling Broadway.Message.put_batch_key/2. Ordering and Partitioning - Broadway allows developers to partition messages across workers, guaranteeing messages within the same partition are processed in order. For example, if you want to guarantee all events tied to a given user_id are processed in order and not concurrently, you can set the :partition_by option. See &quot;Ordering and partitioning&quot;. Rate limiting: Broadway allows developers to rate limit all producers in a single node by a given number of messages in a time period, allowing developers to easily work sources or sinks that cannot cope with a high number of requests. See the &quot;:rate_limiting&quot; option for producers in start_link/2. Statistics/Metrics - Broadway uses the :telemetry library for instrumentation, see &quot;Telemetry&quot; section below for more information. Back-off (TODO) The Broadway Behaviour In order to use Broadway, you need to: Define your pipeline configuration Define a module implementing the Broadway behaviour Example Broadway is a process-based behaviour, and you begin by defining a module that invokes use Broadway. Processes defined by these modules will often be started by a supervisor, and so a start_link/1 function is frequently also defined but not strictly necessary. defmodule MyBroadway do use Broadway def start_link(_opts) do Broadway.start_link(MyBroadway, name: MyBroadwayExample, producer: [ module: {Counter, []}, concurrency: 1 ], processors: [ default: [concurrency: 2] ] ) end ...callbacks... end Then add your Broadway pipeline to your supervision tree (usually in lib/my_app/application.ex): children = [ {MyBroadway, []} ] Supervisor.start_link(children, strategy: :one_for_one) Adding your pipeline to your supervision tree in this way calls the default child_spec/1 function that is generated when use Broadway is invoked. If you would like to customize the child spec passed to the supervisor, you can override the child_spec/1 function in your module or explicitly pass a child spec to the supervisor when adding it to your supervision tree. The configuration above defines a pipeline with: One producer Two processors Here is how this pipeline would be represented: [producer_1] / \\ / \\ / \\ / \\ [processor_1] [processor_2] &lt;- process each message After the pipeline is defined, you need to implement the handle_message/3 callback which will be invoked by processors for each message. handle_message/3 receives every message as a Broadway.Message struct and it must return an updated message. Batching Depending on the scenario, you may want to group processed messages as batches before publishing your data. This is common and especially important when working with services like AWS S3 and SQS that provide a specific API for sending and retrieving batches. This can drastically increase throughput and consequently improve the overall performance of your pipeline. To create batches, define the batchers configuration option: defmodule MyBroadway do use Broadway def start_link(_opts) do Broadway.start_link(MyBroadway, name: MyBroadwayExample, producer: [ module: {Counter, []}, concurrency: 1 ], processors: [ default: [concurrency: 2] ], batchers: [ sqs: [concurrency: 2, batch_size: 10], s3: [concurrency: 1, batch_size: 10] ] ) end ...callbacks... end The configuration above defines a pipeline with: One producer Two processors One batcher named :sqs with two batch processors One batcher named :s3 with one batch processor Here is how this pipeline would be represented: [producer_1] / \\ / \\ / \\ / \\ [processor_1] [processor_2] &lt;- process each message /\\ /\\ / \\ / \\ / \\ / \\ / x \\ / / \\ \\ / / \\ \\ / / \\ \\ [batcher_sqs] [batcher_s3] /\\ \\ / \\ \\ / \\ \\ / \\ \\ [batch_sqs_1] [batch_sqs_2] [batch_s3_1] &lt;- process each batch Additionally, define the handle_batch/4 callback, which batch processors invoke for each batch. You can then call Broadway.Message.put_batcher/2 inside handle_message/3 to control which batcher the message should go to. The batcher receives processed messages and creates batches specified by the batch_size and batch_timeout configuration. The goal is to create a batch with at most batch_size entries within batch_timeout milliseconds. Each message goes into a particular batch, controlled by calling Broadway.Message.put_batch_key/2 in handle_message/3. Once a batch is created, it is sent to a separate process that will call handle_batch/4, passing the batcher, the batch itself (i.e. a list of messages), a Broadway.BatchInfo struct and the Broadway context. For example, imagine your producer generates integers as data. You want to route the odd integers to SQS and the even ones to S3. Your pipeline would look like this: defmodule MyBroadway do use Broadway import Integer alias Broadway.Message ...start_link... @impl true def handle_message(_, %Message{data: data} = message, _) when is_odd(data) do message |&gt; Message.update_data(&amp;process_data/1) |&gt; Message.put_batcher(:sqs) end def handle_message(_, %Message{data: data} = message, _) when is_even(data) do message |&gt; Message.update_data(&amp;process_data/1) |&gt; Message.put_batcher(:s3) end defp process_data(data) do # Do some calculations, generate a JSON representation, etc. end @impl true def handle_batch(:sqs, messages, _batch_info, _context) do # Send batch of successful messages as ACKs to SQS # This tells SQS that this list of messages were successfully processed end def handle_batch(:s3, messages, _batch_info, _context) do # Send batch of messages to S3 end end See the callbacks documentation for more information on the arguments given to each callback and their expected return types. Now you are ready to get started. See the start_link/2 function for a complete reference on the arguments and options allowed. Also makes sure to check out GUIDES in the documentation sidebar for more examples, how tos and more. Acknowledgements and failures At the end of the pipeline, messages are automatically acknowledged. If there are no batchers, the acknowledgement will be done by processors. The number of messages acknowledged, assuming the pipeline is running at full scale, will be max_demand - min_demand. Since the default values are 10 and 5 respectively, we will be acknowledging in groups of 5. If there are batchers, the acknowledgement is done by the batchers, using the batch_size. In case of failures, Broadway does its best to keep the failures contained and avoid losing messages. The failed message or batch is acknowledged as failed immediately. For every failure, a log report is also emitted. If your Broadway module also defines the handle_failed/2 callback, that callback will be invoked with all the failed messages before they get acknowledged. Note however, that Broadway does not provide any sort of retries out of the box. This is left completely as a responsibility of the producer. For instance, if you are using Amazon SQS, the default behaviour is to retry unacknowledged messages after a user-defined timeout. If you don&#39;t want unacknowledged messages to be retried, is your responsibility to configure a dead-letter queue as target for those messages. Testing Many producers receive data from external systems and hitting the network is usually undesirable when running the tests. One way to solve this issue would be to not start Broadway pipeline in tests. Another way would be to use a different producer in tests, one that doesn&#39;t do anything, and that is exactly what Broadway.DummyProducer is for. If the dummy producer doesn&#39;t produce any work, how to test that the pipeline is correct? For that, Broadway ships with a test_messages/2 function. With test_messages/2, you can push some sample data into the pipeline and receive a process message when the pipeline acknowledges the data you have pushed has been processed. This is very useful as a synchronization mechanism. Because many pipelines end-up working with side-effects, you can use the test message acknowledgment to guarantee the message has been processed and therefore side-effects should be visible. For example, if you have a pipeline named MyApp.Broadway that writes to the database on every message, you could test it as: # Push 3 messages with the data field set to 1, 2, and 3 respectively ref = Broadway.test_messages(MyApp.Broadway, [1, 2, 3]) # Assert that the messages have been consumed assert_receive {:ack, ^ref, [_, _, _] = _successful, failed}, 1000 # Now assert the database side-effects ... Also note how we have increased the assert_receive timeout to 1000ms. The default timeout is 100ms, which may not be enough for some pipelines. You may also increase the assert_receive timeout for the whole suite in your test/test_helper.exs: ExUnit.configure(assert_receive_timeout: 2000) When testing pipelines with batchers there are additional considerations. By default, the batch is only delivered when either its size or its timeout has been reached, but that is often impractical for testing, you may not necessarily want to send a lot of data or wait a lot of time for the batch to flush. For this reason, when using test_messages/2, the messages have their :batch_mode set to :flush, causing the batch to be delivered immediately, without waiting for the batch size or the timeout. Ordering and partitioning By default, Broadway processes all messages and batches concurrently, which means ordering is not guaranteed. Some producers may impose some ordering (for instance, Apache Kafka), but if the ordering comes from a business requirement, you will have to impose the ordering yourself. This can be done with the :partition_by option, which enforces that messages with a given property are always forwarded to the same stage. In order to provide partitioning throughout the whole pipeline, just set :partition_by at the root of your configuration: defmodule MyBroadway do use Broadway def start_link(_opts) do Broadway.start_link(MyBroadway, name: MyBroadwayExample, producer: [ module: {Counter, []}, concurrency: 1 ], processors: [ default: [concurrency: 2] ], batchers: [ sqs: [concurrency: 2, batch_size: 10], s3: [concurrency: 1, batch_size: 10] ], partition_by: &amp;partition/1 ) end defp partition(msg) do msg.data.user_id end In the example above, we are partitioning the pipeline by user_id. This means any message with the same user_id will be handled by the same processor and batch processor. The partition function must return a non-negative integer, starting at zero, which is routed to a stage by using the remainder option. If the data you want to partition by is not an integer, you can explicitly hash it by calling :erlang.phash2/1. However, note that hash does not guarantee an equal distribution of events across partitions. So some partitions may be more overloaded than others, slowing down the whole pipeline. In the example above, we have set the same partition for all processors and batchers. You can also specify the :partition_by function for each &quot;processor&quot; and &quot;batcher&quot; individually. Finally, beware of the error semantics when using partitioning. If you require ordering and a message fails, the partition will continue processing messages. Depending on the type of processing, the end result may be inconsistent. If your producer supports retrying, the failed message may be retried later, also out of order. Those issues happens regardless of Broadway and solutions to said problems almost always need to be addressed outside of Broadway too. Telemetry Broadway currently exposes following Telemetry events: [:broadway, :processor, :start] - Dispatched by a Broadway processor when the internal GenStage handle_events/3 callback is invoked Measurement: %{time: System.monotonic_time} Metadata: %{name: atom, messages: [Broadway.Message.t]} [:broadway, :processor, :stop] - Dispatched by a Broadway processor when the internal GenStage handle_events/3 callback has completed processing all the individual messages Measurement: %{time: System.monotonic_time, duration: native_time} Metadata: %{ name: atom, successful_messages_to_ack: [Broadway.Message.t], successful_messages_to_forward: [Broadway.Message.t], failed_messages: [Broadway.Message.t] } [:broadway, :processor, :message, :start] - Dispatched by a Broadway processor before your handle_message/3 callback is invoked Measurement: %{time: System.monotonic_time} Metadata: %{ processor_key: atom, name: atom, message: Broadway.Message.t } [:broadway, :processor, :message, :stop] - Dispatched by a Broadway processor after your handle_message/3 callback has returned Measurement: %{time: System.monotonic_time, duration: native_time} Metadata: %{ processor_key: atom, name: atom, message: Broadway.Message.t, updated_message: Broadway.Message.t } [:broadway, :processor, :message, :error] - Dispatched by a Broadway processor if your handle_message/3 callback encounters an error Measurement: %{time: System.monotonic_time, duration: native_time} Metadata: %{ processor_key: atom, name: atom, message: Broadway.Message.t, kind: kind, reason: reason, stacktrace: stacktrace } [:broadway, :consumer, :start] - Dispatched by a Broadway consumer before your handle_batch/4 callback is invoked Measurement: %{time: System.monotonic_time} Metadata: %{ name: atom, messages: [Broadway.Message.t], batch_info: Broadway.BatchInfo.t } [:broadway, :consumer, :stop] - Dispatched by a Broadway consumer after your handle_batch/4 callback has returned Measurement: %{time: System.monotonic_time, duration: native_time} Metadata: %{ name: atom, successful_messages: [Broadway.Message.t], failed_messages: [Broadway.Message.t], batch_info: Broadway.BatchInfo.t } [:broadway, :batcher, :start] - Dispatched by a Broadway batcher before handling events Measurement: %{time: System.monotonic_time} Metadata: %{name: atom, events: [{Broadway.Message.t}]} [:broadway, :batcher, :stop] - Dispatched by a Broadway batcher after handling events Measurement: %{time: System.monotonic_time, duration: native_time} Metadata: %{name: atom}","ref":"Broadway.html","title":"Broadway","type":"behaviour"},{"doc":"Gets the current values used for the producer rate limiting of the given pipeline. Returns {:ok, info} if rate limiting is enabled for the given pipeline or {:error, reason} if the given pipeline doesn&#39;t have rate limiting enabled. The returned info is a map with the following keys: :interval :allowed_messages See the :rate_limiting options in the module documentation for more information. Examples Broadway.get_rate_limiting(broadway) #=&gt; {:ok, %{allowed_messages: 2000, interval: 1000}}","ref":"Broadway.html#get_rate_limiting/1","title":"Broadway.get_rate_limiting/1","type":"function"},{"doc":"Invoked to handle generated batches. It expects: batcher is the key that defined the batcher. This value can be set in the handle_message/3 callback using Broadway.Message.put_batcher/2. messages is the list of Broadway.Message structs in the incoming batch. batch_info is a Broadway.BatchInfo struct containing extra information about the incoming batch. context is the user defined data structure passed to start_link/2. It must return an updated list of messages. All messages received must be returned, otherwise an error will be logged. All messages after this step will be acknowledged acccording to their status. In case of errors in this callback, the error will be logged and the whole batch will be failed. This callback also traps exits, so failures due to broken links between processes do not automatically cascade.","ref":"Broadway.html#c:handle_batch/4","title":"Broadway.handle_batch/4","type":"callback"},{"doc":"Invoked for failed messages (if defined). It expects: messages is the list of messages that failed. If a message is failed in handle_message/3, this will be a list with a single message in it. If some messages are failed in handle_batch/4, this will be the list of failed messages. context is the user-defined data structure passed to start_link/2. This callback must return the same messages given to it, possibly updated. For example, you could update the message data or use Broadway.Message.configure_ack/2 in a centralized place to configure how to ack the message based on the failure reason. This callback is optional. If present, it&#39;s called before the messages are acknowledged according to the producer. This gives you a chance to do something with the message before it&#39;s acknowledged, such as storing it in an external persistence layer or similar. This callback is also invoked if handle_message/3 or handle_batch/4 crash or raise an error. If this callback crashes or raises an error, the messages are failed internally by Broadway to avoid crashing the process.","ref":"Broadway.html#c:handle_failed/2","title":"Broadway.handle_failed/2","type":"callback"},{"doc":"Invoked to handle/process individual messages sent from a producer. It receives: processor is the key that defined the processor. message is the Broadway.Message struct to be processed. context is the user defined data structure passed to start_link/2. And it must return the (potentially) updated Broadway.Message struct. This is the place to do any kind of processing with the incoming message, e.g., transform the data into another data structure, call specific business logic to do calculations. Basically, any CPU bounded task that runs against a single message should be processed here. In order to update the data after processing, use the Broadway.Message.update_data/2 function. This way the new message can be properly forwarded and handled by the batcher: @impl true def handle_message(_, message, _) do message |&gt; update_data(&amp;do_calculation_and_returns_the_new_data/1) end In case more than one batcher have been defined in the configuration, you need to specify which of them the resulting message will be forwarded to. You can do this by calling put_batcher/2 and returning the new updated message: @impl true def handle_message(_, message, _) do # Do whatever you need with the data ... message |&gt; put_batcher(:s3) end Any message that has not been explicitly failed will be forwarded to the next step in the pipeline. If there are no extra steps, it will be automatically acknowledged. In case of errors in this callback, the error will be logged and that particular message will be immediately acknowledged as failed, not proceeding to the next steps of the pipeline. This callback also traps exits, so failures due to broken links between processes do not automatically cascade.","ref":"Broadway.html#c:handle_message/3","title":"Broadway.handle_message/3","type":"callback"},{"doc":"Returns the names of producers. Examples iex&gt; Broadway.producer_names(MyBroadway) [MyBroadway.Producer_0, MyBroadway.Producer_1, ..., MyBroadway.Producer_7]","ref":"Broadway.html#producer_names/1","title":"Broadway.producer_names/1","type":"function"},{"doc":"Sends a list of Broadway.Messages to the Broadway pipeline. The producer is randomly chosen among all sets of producers/stages. This is used to send out of band data to a Broadway pipeline.","ref":"Broadway.html#push_messages/2","title":"Broadway.push_messages/2","type":"function"},{"doc":"Starts a Broadway process linked to the current process. module is the module implementing the Broadway behaviour. Options In order to set up how the pipeline created by Broadway should work, you need to specify the blueprint of the pipeline. You can do this by passing a set of options to start_link/2. Each component of the pipeline has its own set of options. The broadway options are: :name - Required. Used for name registration. All processes/stages created will be named using this value as prefix. :producer - Required. A keyword list of options. See &quot;Producers options&quot; section below. Only a single producer is allowed. :processors - Required. A keyword list of named processors where the key is an atom as identifier and the value is another keyword list of options. See &quot;Processors options&quot; section below. Currently only a single processor is allowed. :batchers - Optional. A keyword list of named batchers where the key is an atom as identifier and the value is another keyword list of options. See &quot;Batchers options&quot; section below. :context - Optional. A user defined data structure that will be passed to handle_message/3 and handle_batch/4. :shutdown - Optional. The time in milliseconds given for Broadway to gracefully shutdown without discarding events. Defaults to 30_000(ms). :resubscribe_interval - Optional. The interval in milliseconds that processors wait until they resubscribe to a failed producers. Defaults to 100(ms). :partition_by - Optional. A function that controls how data is partitioned across all processors and batchers. It receives a Broadway.Message and it must return a non-negative integer, starting with zero, that will be mapped to one of the existing processors. See &quot;Ordering and Partitioning&quot; in the module docs for more information. :hibernate_after - Optional. If a process does not receive any message within this interval, it will hibernate, compacting memory. Applies to producers, processors, and batchers. Defaults to 15_000(ms). :spawn_opt - Optional. Low-level options given when starting a process. Applies to producers, processors, and batchers. See erlang:spawn_opt/2 for more information. Producers options The producer options are: :module - Required. A tuple representing a GenStage producer. The tuple format should be {mod, arg}, where mod is the module that implements the GenStage behaviour and arg the argument that will be passed to the init/1 callback of the producer. Pay attention that this producer must emit events that are Broadway.Message structs. It&#39;s recommended that arg is a keyword list. In fact, if arg is a keyword list, a :broadway option is injected into such keyword list containing the configuration for the complete Broadway topology with the addition of an :index key, telling the index of the producer in its supervision tree (starting from 0). This allows a features such having even producers connect to some server while odd producers connect to another. :concurrency - Optional. The number of concurrent producers that will be started by Broadway. Use this option to control the concurrency level of each set of producers. The default value is 1. :transformer - Optional. A tuple representing a transformer that translates a produced GenStage event into a %Broadway.Message{}. The tuple format should be {mod, fun, opts} and the function should have the following spec (event :: term, opts :: term) :: Broadway.Message.t This function must be used sparingly and exclusively to convert regular messages into Broadway.Message. That&#39;s because a failure in the :transformer callback will cause the whole producer to terminate, possibly leaving unacknowledged messages along the way. :rate_limiting - Optional. A list of options to enable and configure rate limiting for producing. If this option is present, rate limiting is enabled, otherwise it isn&#39;t. Rate limiting refers to the rate at which producers will forward messages to the rest of the pipeline. The rate limiting is applied to and shared by all producers within the time limit. The following options are supported: :allowed_messages - Required. An integer that describes how many messages are allowed in the specified interval. :interval - Required. An integer that describes the interval (in milliseconds) during which the number of allowed messages is allowed. If the producer produces more than allowed_messages in interval, only allowed_messages will be published until the end of interval, and then more messages will be published. :hibernate_after - Optional. Overrides the top-level :hibernate_after. :spawn_opt - Optional. Overrides the top-level :spawn_opt. Processors options The processors options are: :concurrency - Optional. The number of concurrent process that will be started by Broadway. Use this option to control the concurrency level of the processors. The default value is System.schedulers_online() * 2. :min_demand - Optional. Set the minimum demand of all processors stages. Default value is 5. :max_demand - Optional. Set the maximum demand of all processors stages. Default value is 10. :partition_by - Optional. Overrides the top-level :partition_by. :hibernate_after - Optional. Overrides the top-level :hibernate_after. :spawn_opt - Optional. Overrides the top-level :spawn_opt. Batchers options :concurrency - Optional. The number of concurrent batch processors that will be started by Broadway. Use this option to control the concurrency level. Note that this only sets the numbers of batch processors for each batcher group, not the number of batchers. The number of batchers will always be one for each batcher key defined. The default value is 1. :batch_size - Optional. The size of the generated batches. Default value is 100. :batch_timeout - Optional. The time, in milliseconds, that the batcher waits before flushing the list of messages. When this timeout is reached, a new batch is generated and sent downstream, no matter if the :batch_size has been reached or not. Default value is 1000 (1 second). :partition_by - Optional. Overrides the top-level :partition_by. :hibernate_after - Optional. Overrides the top-level :hibernate_after. :spawn_opt - Optional. Overrides the top-level :spawn_opt.","ref":"Broadway.html#start_link/2","title":"Broadway.start_link/2","type":"function"},{"doc":"Sends a list of data as messages to the Broadway pipeline. This is a convenience used mostly for testing. The given data is automatically wrapped in a Broadway.Message with Broadway.CallerAcknowledger configured to send a message back to the caller once the message has been fully processed. It uses push_messages/2 for dispatching. It returns a reference that can be used to identify the ack messages. See &quot;Testing&quot; section in module documentation for more information. Options :batch_mode - when set to :flush, the batch the message is in is immediately delivered. When set to :bulk, batch is delivered when its size or timeout is reached. Defaults to :flush. :metadata - optionally a map of additional fields to add to the message. This can be used, for example, when testing BroadwayRabbitMQ.Producer. Examples For example, in your tests, you may do: ref = Broadway.test_messages(broadway, [1, 2, 3]) assert_receive {:ack, ^ref, successful, failed} assert length(successful) == 3 assert length(failed) == 0","ref":"Broadway.html#test_messages/3","title":"Broadway.test_messages/3","type":"function"},{"doc":"Updates the producer rate limiting of the given pipeline at runtime. Supports the following options (see the :rate_limiting options in the module documentation for more information): :allowed_messages :interval Returns an {:error, reason} tuple if the given broadway pipeline doesn&#39;t have rate limiting enabled. Examples Broadway.update_rate_limiting(broadway, allowed_messages: 100)","ref":"Broadway.html#update_rate_limiting/2","title":"Broadway.update_rate_limiting/2","type":"function"},{"doc":"A struct used to hold information about a generated batch. An instance of this struct containing the related info will be passed to the Broadway.handle_batch/4 callback of the module implementing the Broadway behaviour.","ref":"Broadway.BatchInfo.html","title":"Broadway.BatchInfo","type":"module"},{"doc":"","ref":"Broadway.BatchInfo.html#t:t/0","title":"Broadway.BatchInfo.t/0","type":"type"},{"doc":"This struct holds all information about a message. A message is first created by the producers. It is then sent downstream and gets updated multiple times, either by a module implementing the Broadway behaviour through the Broadway.handle_message/3 callback or internally by one of the built-in stages of Broadway. Instead of modifying the struct directly, you should use the functions provided by this module to manipulate messages.","ref":"Broadway.Message.html","title":"Broadway.Message","type":"module"},{"doc":"Immediately acknowledges the given message or list of messages. This function can be used to acknowledge a message (or list of messages) immediately without waiting for the rest of the pipeline. Acknowledging a message sets that message&#39;s acknowledger to a no-op acknowledger so that it&#39;s safe to ack at the end of the pipeline. Returns the updated acked message if a message is passed in, or the updated list of acked messages if a list of messages is passed in.","ref":"Broadway.Message.html#ack_immediately/1","title":"Broadway.Message.ack_immediately/1","type":"function"},{"doc":"Configures the acknowledger of this message. This function calls the Broadway.Acknowledger.configure/3 callback to change the configuration of the acknowledger for the given message. This function can only be called if the acknowledger implements the configure/3 callback. If it doesn&#39;t, an error is raised.","ref":"Broadway.Message.html#configure_ack/2","title":"Broadway.Message.configure_ack/2","type":"function"},{"doc":"Mark a message as failed. Failed messages are sent directly to the related acknowledger so they&#39;re not forwarded to the next step in the pipeline.","ref":"Broadway.Message.html#failed/2","title":"Broadway.Message.failed/2","type":"function"},{"doc":"Defines the message batch key. Batcher functions then attempt to create batches with the same batch_key, of size batch_size within period batch_timeout.","ref":"Broadway.Message.html#put_batch_key/2","title":"Broadway.Message.put_batch_key/2","type":"function"},{"doc":"Sets the batching mode for the message. When the mode is :bulk, the batch that the message is in is delivered after the batch size or batch timeout is reached. When the mode is :flush, the batch that the message is in is delivered immediately after processing. Note it doesn&#39;t mean the batch contains only a single element but rather that all messages received from the processor are delivered without waiting. The default mode for messages is :bulk.","ref":"Broadway.Message.html#put_batch_mode/2","title":"Broadway.Message.put_batch_mode/2","type":"function"},{"doc":"Defines the target batcher which the message should be forwarded to.","ref":"Broadway.Message.html#put_batcher/2","title":"Broadway.Message.put_batcher/2","type":"function"},{"doc":"Updates the data in a message. This function is usually used inside the Broadway.handle_message/3 implementation to replace data with new processed data.","ref":"Broadway.Message.html#update_data/2","title":"Broadway.Message.update_data/2","type":"function"},{"doc":"","ref":"Broadway.Message.html#t:t/0","title":"Broadway.Message.t/0","type":"type"},{"doc":"A behaviour used to acknowledge that the received messages were successfully processed or failed. When implementing a new connector for Broadway, you should implement this behaviour and consider how the technology you&#39;re working with handles message acknowledgement. The ack/3 callback must be implemented in order to notify the origin of the data that a message can be safely removed after been successfully processed and published. In case of failed messages or messages without acknowledgement, depending on the technology chosen, the messages can be either moved back in the queue or, alternatively, moved to a dead-letter queue.","ref":"Broadway.Acknowledger.html","title":"Broadway.Acknowledger","type":"behaviour"},{"doc":"Invoked to acknowledge successful and failed messages. ack_ref is a term that uniquely identifies how messages should be grouped and sent for acknowledgement. Imagine you have a scenario where messages are coming from different producers. Broadway will use this information to correctly identify the acknowledger and pass it among with the messages so you can properly communicate with the source of the data for acknowledgement. successful is the list of messages that were successfully processed and published. failed is the list of messages that, for some reason, could not be processed or published.","ref":"Broadway.Acknowledger.html#c:ack/3","title":"Broadway.Acknowledger.ack/3","type":"callback"},{"doc":"Configures the acknowledger with new options. Every acknowledger can decide how to incorporate the given options into its ack_data. The ack_data is the current acknowledger&#39;s data. The return value of this function is {:ok, new_ack_data} where new_ack_data is the updated data for the acknowledger. Note that options are different for every acknowledger, as the acknowledger is what specifies what are the supported options. Check the documentation for the acknowledger you&#39;re using to see the supported options.","ref":"Broadway.Acknowledger.html#c:configure/3","title":"Broadway.Acknowledger.configure/3","type":"callback"},{"doc":"A simple acknowledger that sends a message back to a caller. It must be stored as: acknowledger: {Broadway.CallerAcknowledger, {pid, ref}, term} The second element is a tuple with the pid to receive the messages and a unique identifier (usually a reference). The third element, which is per message, is ignored. It sends a message in the format: {:ack, ref, successful_messages, failed_messages} If Broadway.Message.configure_ack/2 is called on a message that uses this acknowledger, then the following message is sent: {:configure, ref, options}","ref":"Broadway.CallerAcknowledger.html","title":"Broadway.CallerAcknowledger","type":"module"},{"doc":"An acknowledger that does nothing. Set automatically on messages that have been acked immediately via Broadway.Message.ack_immediately/1.","ref":"Broadway.NoopAcknowledger.html","title":"Broadway.NoopAcknowledger","type":"module"},{"doc":"A producer that does nothing, used mostly for testing. See &quot;Testing&quot; section in Broadway module documentation for more information.","ref":"Broadway.DummyProducer.html","title":"Broadway.DummyProducer","type":"module"},{"doc":"A simple term storage to avoid passing large amounts of data between processes. If you have a large amount of data and you want to avoid passing it between processes, you can use the TermStorage. The TermStorage creates a unique reference for it, allowing you pass the reference around instead of the term. If the same term is put multiple times, it is stored only once, avoiding generating garbage. However, the stored terms are never removed. A common use case for this feature is in Broadway.Producer, which may need to pass a lot of information to acknowledge messages. With this module, you can store those terms when the producer starts and only pass the reference between messages: iex&gt; ref = Broadway.TermStorage.put({:foo, :bar, :baz}) # On init iex&gt; Broadway.TermStorage.get!(ref) # On ack {:foo, :bar, :baz}","ref":"Broadway.TermStorage.html","title":"Broadway.TermStorage","type":"module"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor.","ref":"Broadway.TermStorage.html#child_spec/1","title":"Broadway.TermStorage.child_spec/1","type":"function"},{"doc":"Gets a previously stored term.","ref":"Broadway.TermStorage.html#get!/1","title":"Broadway.TermStorage.get!/1","type":"function"},{"doc":"Puts a term.","ref":"Broadway.TermStorage.html#put/1","title":"Broadway.TermStorage.put/1","type":"function"},{"doc":"Amazon SQS Amazon Simple Queue Service (SQS) is a highly scalable distributed message queuing service provided by Amazon.com. AWS SQS offers two types of message queues: Standard Nearly unlimited throughput Best-effort ordering At-least-once delivery FIFO Limited number of transactions per second (TPS). See Amazon SQS FIFO developer guide for more information on limits. Order in which messages are sent/received is strictly preserved Exactly-once delivery Broadway can work seamlessly with both, Standard and FIFO queues.","ref":"amazon-sqs.html","title":"Amazon SQS","type":"extras"},{"doc":"In order to use Broadway with SQS, we need to: Create a SQS queue (or use an existing one) Configure our Elixir project to use Broadway Define your pipeline configuration Implement Broadway callbacks Run the Broadway pipeline Tuning the configuration (Optional)","ref":"amazon-sqs.html#getting-started","title":"Amazon SQS - Getting Started","type":"extras"},{"doc":"Amazon provides a comprehensive Step-by-step Guide on creating SQS queues. In case you don&#39;t have an AWS account and want to test Broadway locally, use can easily install ElasticMQ, which is a message queue system that offers a SQS-compatible query interface.","ref":"amazon-sqs.html#create-a-sqs-queue","title":"Amazon SQS - Create a SQS queue","type":"extras"},{"doc":"In this guide we&#39;re going to use BroadwaySQS, which is a Broadway SQS Connector provided by Dashbit. Starting a new project If you plan to start a new project, just run: mix new my_app --sup The --sup flag instructs Elixir to generate an application with a supervision tree. Setting up dependencies Add :broadway_sqs to the list of dependencies in mix.exs along the HTTP client of your choice (defaults to :hackney): def deps do [ ... {:broadway_sqs, &quot;~&gt; 0.2.0&quot;}, {:hackney, &quot;~&gt; 1.9&quot;}, ] end Don&#39;t forget to check for the latest version of dependencies.","ref":"amazon-sqs.html#configure-the-project","title":"Amazon SQS - Configure the project","type":"extras"},{"doc":"Broadway is a process-based behaviour and to define a Broadway pipeline, we need to define three functions: start_link/1, handle_message/3 and handle_batch/4. We will cover start_link/1 in this section and the handle_ callbacks in the next one. Similar to other process-based behaviour, start_link/1 simply delegates to Broadway.start_link/2, which should define the producers, processors, and batchers in the Broadway pipeline. Assuming we want to consume messages from a queue called my_queue, the minimal configuration would be: defmodule MyBroadway do use Broadway alias Broadway.Message def start_link(_opts) do Broadway.start_link(__MODULE__, name: __MODULE__, producer: [ module: {BroadwaySQS.Producer, queue_url: &quot;https://us-east-2.queue.amazonaws.com/100000000001/my_queue&quot;} ], processors: [ default: [] ], batchers: [ default: [ batch_size: 10, batch_timeout: 2000 ] ] ) end ...callbacks... end The above configuration also assumes that you have the AWS credentials set up in your environment, for instance, by having the AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY environment variables set. If that&#39;s not the case, you will need to pass that information to the client so it can properly connect to the AWS servers. Here is how you can do it: ... producer: [ module: {BroadwaySQS.Producer, queue_url: &quot;https://us-east-2.queue.amazonaws.com/100000000001/my_queue&quot;, config: [ access_key_id: &quot;YOUR_AWS_ACCESS_KEY_ID&quot;, secret_access_key: &quot;YOUR_AWS_SECRET_ACCESS_KEY&quot; ] } ] ... For a full list of options for BroadwaySQS.Producer, please see BroadwaySQS documentation. For general information about setting up Broadway, see Broadway module docs as well as Broadway.start_link/2. Note: Even though batching is optional since Broadway v0.2, we recommend that all Amazon SQS pipelines have at least a default batcher. This lets you control the exact batch size and frequency that messages are acknowledged to Amazon SQS, often leading to pipelines that are more cost and time efficient.","ref":"amazon-sqs.html#define-the-pipeline-configuration","title":"Amazon SQS - Define the pipeline configuration","type":"extras"},{"doc":"In order to process incoming messages, we need to implement the required callbacks. For the sake of simplicity, we&#39;re considering that all messages received from the queue are just numbers: defmodule MyBroadway do use Broadway alias Broadway.Message ...start_link... @impl true def handle_message(_, %Message{data: data} = message, _) do message |&gt; Message.update_data(fn data -&gt; data * data end) end @impl true def handle_batch(_, messages, _, _) do list = messages |&gt; Enum.map(fn e -&gt; e.data end) IO.inspect(list, label: &quot;Got batch of finished jobs from processors, sending ACKs to SQS as a batch.&quot;) messages end end We are not doing anything fancy here, but it should be enough for our purpose. First we update the message&#39;s data individually inside handle_message/3 and then we print each batch inside handle_batch/4. For more information, see Broadway.handle_message/3 and Broadway.handle_batch/4.","ref":"amazon-sqs.html#implement-broadway-callbacks","title":"Amazon SQS - Implement Broadway callbacks","type":"extras"},{"doc":"To run your Broadway pipeline, you just need to add as a child in a supervision tree. Most applications have a supervision tree defined at lib/my_app/application.ex. You can add Broadway as a child to a supervisor as follows: children = [ {MyBroadway, []} ] Supervisor.start_link(children, strategy: :one_for_one) Now the Broadway pipeline should be started when your application starts. Also, if your Broadway has any dependency (for example, it needs to talk to the database), make sure that Broadway is listed after its dependencies in the supervision tree.","ref":"amazon-sqs.html#run-the-broadway-pipeline","title":"Amazon SQS - Run the Broadway pipeline","type":"extras"},{"doc":"Some of the configuration options available for Broadway come already with a &quot;reasonable&quot; default value. However those values might not suit your requirements. Depending on the number of messages you get, how much processing they need and how much IO work is going to take place, you might need completely different values to optimize the flow of your pipeline. The stages option available for every set of producers, processors and batchers, among with batch_size and batch_timeout can give you a great deal of flexibility. The stages option controls the concurrency level in each layer of the pipeline. Here&#39;s an example on how you could tune them according to your needs. defmodule MyBroadway do use Broadway def start_link(_opts) do Broadway.start_link(__MODULE__, name: __MODULE__, producer: [ ... concurrency: 60, ], processors: [ default: [ concurrency: 100, ] ], batchers: [ default: [ batch_size: 10, concurrency: 80, ] ] ) end ...callbacks... end In order to get a good set of configurations for your pipeline, it&#39;s important to respect the limitations of the servers you&#39;re running, as well as the limitations of the services you&#39;re providing/consuming data to/from.","ref":"amazon-sqs.html#tuning-the-configuration","title":"Amazon SQS - Tuning the configuration","type":"extras"},{"doc":"Google Cloud Pub/Sub Cloud Pub/Sub is a fully-managed real-time messaging service provided by Google.","ref":"google-cloud-pubsub.html","title":"Google Cloud Pub/Sub","type":"extras"},{"doc":"In order to use Broadway with Cloud Pub/Sub you need to: Setup a Cloud Pub/Sub project Configure your Elixir project to use Broadway Define your pipeline configuration Implement Broadway callbacks Run the Broadway pipeline Tune the configuration (Optional) If you are just getting familiar with Google Pub/Sub, refer to the documentation to get started. Instead of testing against a live environment, you may also consider using the emulator to simulate integrating with Cloud Pub/Sub. If you have an existing project, topic, subscription, and credentials, you can skip step 1 and jump to Configure the project section.","ref":"google-cloud-pubsub.html#getting-started","title":"Google Cloud Pub/Sub - Getting Started","type":"extras"},{"doc":"In this tutorial we&#39;ll use the gcloud command-line tool to set everything up in Google Cloud. Alternatively, you can roughly follow this guide by using Cloud Console. To install gcloud follow the documentation. If you are on macOS you may consider installing it with Homebrew: brew cask install google-cloud-sdk Now, authenticate the CLI: gcloud auth login Then, create a new project: gcloud projects create test-pubsub A new topic: gcloud pubsub topics create test-topic --project test-pubsub Created topic [projects/test-pubsub/topics/test-topic]. Note: If you run this command immediately after creating a new Google Cloud project, you may receive an error indicating that your project&#39;s organization policy is still being provisioned. Just wait a couple minutes and try again. And a new subscription: gcloud pubsub subscriptions create test-subscription --project test-pubsub --topic test-topic Created subscription [projects/test-pubsub/subscriptions/test-subscription]. We also need a service account, an IAM policy, as well as API credentials in order to programmatically work with the service. First, let&#39;s create the service account: gcloud iam service-accounts create test-account --project test-pubsub Created service account [test-account]. Then the policy. For simplicity we add the general role roles/editor, but make sure to examine the available roles and choose the one that best suits your use case: gcloud projects add-iam-policy-binding test-pubsub \\ --member serviceAccount:test-account@test-pubsub.iam.gserviceaccount.com \\ --role roles/editor Updated IAM policy for project [test-pubsub]. (...) And now the credentials: gcloud iam service-accounts keys create credentials.json --iam-account=test-account@test-pubsub.iam.gserviceaccount.com created key [xxx] of type [json] as [key] for [test-account@test-pubsub.iam.gserviceaccount.com] This command generated a credentials.json file which will be useful later. Note, the IAM account pattern is &lt;account&gt;@&lt;project&gt;.iam.gserviceaccount.com. Run gcloud iam service-accounts list --project test-pubsub to see all service accounts associated with the given project. Finally, we need to enable Pub/Sub for our project: gcloud services enable pubsub --project test-pubsub Operation &quot;operations/xxx&quot; finished successfully.","ref":"google-cloud-pubsub.html#setup-cloud-pub-sub-project","title":"Google Cloud Pub/Sub - Setup Cloud Pub/Sub project","type":"extras"},{"doc":"In this guide we&#39;re going to use BroadwayCloudPubSub, which is a Broadway Cloud Pub/Sub Connector provided by Dashbit. Starting a new project If you plan to start a new project, just run: mix new my_app --sup The --sup flag instructs Elixir to generate an application with a supervision tree. Setting up dependencies Add :broadway_cloud_pub_sub to the list of dependencies in mix.exs, along with the Google Cloud authentication library of your choice (defaults to :goth): defp deps() do [ ... {:broadway_cloud_pub_sub, &quot;~&gt; 0.4.0&quot;}, {:goth, &quot;~&gt; 1.0&quot;} ] end Don&#39;t forget to check for the latest version of dependencies.","ref":"google-cloud-pubsub.html#configure-the-project","title":"Google Cloud Pub/Sub - Configure the project","type":"extras"},{"doc":"Broadway is a process-based behaviour and to define a Broadway pipeline, we need to define three functions: start_link/1, handle_message/3 and handle_batch/4. We will cover start_link/1 in this section and the handle_ callbacks in the next one. Similar to other process-based behaviour, start_link/1 simply delegates to Broadway.start_link/2, which should define the producers, processors, and batchers in the Broadway pipeline. Assuming we want to consume messages from the test-subscription, the minimal configuration would be: defmodule MyBroadway do use Broadway alias Broadway.Message def start_link(_opts) do Broadway.start_link(__MODULE__, name: __MODULE__, producer: [ module: {BroadwayCloudPubSub.Producer, subscription: &quot;projects/test-pubsub/subscriptions/test-subscription&quot;} ], processors: [ default: [] ], batchers: [ default: [ batch_size: 10, batch_timeout: 2_000 ] ] ) end ...callbacks... end For a full list of options for BroadwayCloudPubSub.Producer, please see the documentation. For general information about setting up Broadway, see Broadway module docs as well as Broadway.start_link/2. Note: Even though batching is optional since Broadway v0.2, we recommend all Cloud Pub/Sub pipelines to have at least a default batcher, as that allows you to control the exact batch size and frequency that messages are acknowledged to Cloud Pub/Sub, which often leads to pipelines that are more cost and time efficient.","ref":"google-cloud-pubsub.html#define-the-pipeline-configuration","title":"Google Cloud Pub/Sub - Define the pipeline configuration","type":"extras"},{"doc":"In order to process incoming messages, we need to implement the required callbacks. For the sake of simplicity, we&#39;re considering that all messages received from the queue are strings and our processor calls String.upcase/1 on them: defmodule MyBroadway do use Broadway alias Broadway.Message ...start_link... def handle_message(_, %Message{data: data} = message, _) do message |&gt; Message.update_data(fn data -&gt; String.upcase(data) end) end def handle_batch(_, messages, _, _) do list = messages |&gt; Enum.map(fn e -&gt; e.data end) IO.inspect(list, label: &quot;Got batch of finished jobs from processors, sending ACKs to Pub/Sub as a batch.&quot;) messages end end We are not doing anything fancy here, but it should be enough for our purpose. First we update the message&#39;s data individually inside handle_message/3 and then we print each batch inside handle_batch/4. For more information, see Broadway.handle_message/3 and Broadway.handle_batch/4.","ref":"google-cloud-pubsub.html#implement-broadway-callbacks","title":"Google Cloud Pub/Sub - Implement Broadway callbacks","type":"extras"},{"doc":"To run your Broadway pipeline, you need to add it as a child in a supervision tree. Most applications have a supervision tree defined at lib/my_app/application.ex. You can add Broadway as a child to a supervisor as follows: children = [ {MyBroadway, []} ] Supervisor.start_link(children, strategy: :one_for_one) The final step is to configure credentials. You can set the following environment variable: export GOOGLE_APPLICATION_CREDENTIALS=/path/to/credentials.json See Goth documentation for alternative ways of authenticating with the API. Now the Broadway pipeline should be started when your application starts. Also, if your Broadway pipeline has any dependency (for example, it needs to talk to the database), make sure that it is listed after its dependencies in the supervision tree. If you followed the previous section about setting the project with gcloud, you can now test the the pipeline. In one terminal tab start the application: iex -S mix And in another tab, send a couple of test messages to Pub/Sub: gcloud pubsub topics publish projects/test-pubsub/topics/test-topic --message &quot;test 1&quot; messageIds: - &#39;651428033718119&#39; gcloud pubsub topics publish projects/test-pubsub/topics/test-topic --message &quot;test 2&quot; messageIds: - &#39;651427034966696&#39; Now, In the first tab, you should see output similar to: Got batch of finished jobs from processors, sending ACKs to Pub/Sub as a batch.: [&quot;TEST 1&quot;, &quot;TEST 2&quot;]","ref":"google-cloud-pubsub.html#run-the-broadway-pipeline","title":"Google Cloud Pub/Sub - Run the Broadway pipeline","type":"extras"},{"doc":"Some of the configuration options available for Broadway come already with a &quot;reasonable&quot; default value. However those values might not suit your requirements. Depending on the number of messages you get, how much processing they need and how much IO work is going to take place, you might need completely different values to optimize the flow of your pipeline. The stages option available for every set of producers, processors and batchers, among with batch_size and batch_timeout can give you a great deal of flexibility. The stages option controls the concurrency level in each layer of the pipeline. Here&#39;s an example on how you could tune them according to your needs. defmodule MyBroadway do use Broadway def start_link(_opts) do Broadway.start_link(__MODULE__, name: __MODULE__, producer: [ ... concurrency: 60, ], processors: [ default: [ concurrency: 100, ] ], batchers: [ default: [ batch_size: 10, concurrency: 80, ] ] ) end ...callbacks... end In order to get a good set of configurations for your pipeline, it&#39;s important to respect the limitations of the servers you&#39;re running, as well as the limitations of the services you&#39;re providing/consuming data to/from.","ref":"google-cloud-pubsub.html#tuning-the-configuration","title":"Google Cloud Pub/Sub - Tuning the configuration","type":"extras"},{"doc":"RabbitMQ RabbitMQ is an open source message broker designed to be highly scalable and distributed. It supports multiple protocols including the Advanced Message Queuing Protocol (AMQP).","ref":"rabbitmq.html","title":"RabbitMQ","type":"extras"},{"doc":"In order to use Broadway with RabbitMQ, we need to: Create a queue (or use an existing one) Configure our Elixir project to use Broadway Define your pipeline configuration Implement Broadway callbacks Run the Broadway pipeline Tuning the configuration (Optional) In case you want to work with an existing queue, you can skip step 1 and jump to Configure the project. Note: BroadwayRabbitMQ does not automatically create any queue. If you configure a pipeline with a non-existent queue, the producers will crash, bringing down the pipeline.","ref":"rabbitmq.html#getting-started","title":"RabbitMQ - Getting Started","type":"extras"},{"doc":"RabbitMQ runs on many operating systems. Please see Downloading and Installing RabbitMQ for further information. Also, make sure you have the Management plugin enabled, which ships with the command line tool, rabbitmqadmin. After successfully installing RabbitMQ, you can declare a new queue with the following command: $ rabbitmqadmin declare queue name=my_queue durable=true You can list all declared queues to see our the one we&#39;ve just created: $ rabbitmqctl list_queues Timeout: 60.0 seconds ... Listing queues for vhost / ... name messages my_queue 0","ref":"rabbitmq.html#create-a-queue","title":"RabbitMQ - Create a queue","type":"extras"},{"doc":"In this guide, we&#39;re going to use BroadwayRabbitMQ, which is a Broadway RabbitMQ Connector provided by Dashbit. Starting a new project If you&#39;re creating a new project, run: mix new my_app --sup The --sup flag instructs Elixir to generate an application with a supervision tree. Setting up dependencies Add :broadway_rabbitmq to the list of dependencies in mix.exs: def deps do [ ... {:broadway_rabbitmq, &quot;~&gt; 0.1.0&quot;}, ] end Don&#39;t forget to check for the latest version of dependencies.","ref":"rabbitmq.html#configure-the-project","title":"RabbitMQ - Configure the project","type":"extras"},{"doc":"Broadway is a process-based behaviour and to define a Broadway pipeline, we need to define three functions: start_link/1, handle_message/3 and optionally handle_batch/4. We will cover start_link/1 in this section and the handle_ callbacks in the next one. Similar to other process-based behaviours, start_link/1 simply delegates to Broadway.start_link/2, which should define the producers, processors, and batchers in the Broadway pipeline. Assuming we want to consume messages from a queue called my_queue, one possible configuration would be: defmodule MyBroadway do use Broadway alias Broadway.Message def start_link(_opts) do Broadway.start_link(__MODULE__, name: MyBroadway, producer: [ module: {BroadwayRabbitMQ.Producer, queue: &quot;my_queue&quot;, qos: [ prefetch_count: 50, ] }, concurrency: 2 ], processors: [ default: [ concurrency: 50 ] ], batchers: [ default: [ batch_size: 10, batch_timeout: 1500, concurrency: 5 ] ] ) end ...callbacks... end If you&#39;re consuming data from an existing broker that requires authorization, you&#39;ll need to provide your credentials using the connection option: ... producer: [ module: {BroadwayRabbitMQ.Producer, queue: &quot;my_queue&quot;, connection: [ username: &quot;user&quot;, password: &quot;password&quot;, ] ... } ] ... For the full list of connection options, please see AMQP.Connection.open/1 For general information about setting up Broadway, see Broadway module docs as well as Broadway.start_link/2.","ref":"rabbitmq.html#define-the-pipeline-configuration","title":"RabbitMQ - Define the pipeline configuration","type":"extras"},{"doc":"In order to process incoming messages, we need to implement the required callbacks. For the sake of simplicity, we&#39;re considering that all messages received from the queue are just numbers: defmodule MyBroadway do use Broadway alias Broadway.Message ...start_link... @impl true def handle_message(_, message, _) do message |&gt; Message.update_data(fn data -&gt; {data, String.to_integer(data) * 2} end) end @impl true def handle_batch(_, messages, _, _) do list = messages |&gt; Enum.map(fn e -&gt; e.data end) IO.inspect(list, label: &quot;Got batch&quot;) messages end end We are not doing anything fancy here, but it should be enough for our purpose. First, we update the message&#39;s data individually inside handle_message/3 and then we print each batch inside handle_batch/4. For more information, see Broadway.handle_message/3 and Broadway.handle_batch/4. Note: Since Broadway v0.2, batching is optional. In case you don&#39;t need to group messages as batches for further processing/publishing, you can remove the :batchers configuration along with the handle_batch/4 callback. This is perfectly fine for RabbitMQ, where messages are acknowledged invidivually and never as a batch.","ref":"rabbitmq.html#implement-broadway-callbacks","title":"RabbitMQ - Implement Broadway callbacks","type":"extras"},{"doc":"To run your Broadway pipeline, you just need to add as a child in a supervision tree. Most applications have a supervision tree defined at lib/my_app/application.ex. You can add Broadway as a child to a supervisor as follows: children = [ {MyBroadway, []} ] Supervisor.start_link(children, strategy: :one_for_one) Now the Broadway pipeline should be started when your application starts. Also, if your Broadway has any dependency (for example, it needs to talk to the database), make sure that Broadway is listed after its dependencies in the supervision tree. You can now test your pipeline by entering an iex session: iex -S mix If everything went fine, you should see lots of info log messages from the amqp supervisors. If you think that&#39;s too verbose and want to do something about it, please take a look at the &quot;Log related to amqp supervisors are too verbose&quot; subsection in the amqp&#39;s Troubleshooting documentation. Finally, let&#39;s generate some sample messages to be consumed by Broadway with the following code: {:ok, connection} = AMQP.Connection.open {:ok, channel} = AMQP.Channel.open(connection) AMQP.Queue.declare(channel, &quot;my_queue&quot;, durable: true) Enum.each(1..5000, fn i -&gt; AMQP.Basic.publish(channel, &quot;&quot;, &quot;my_queue&quot;, &quot;\#{i}&quot;) end) AMQP.Connection.close(connection) Use should see the output showing the generated batches: Got batch: [ {&quot;7&quot;, 14}, {&quot;5&quot;, 10}, {&quot;8&quot;, 16}, {&quot;98&quot;, 196}, {&quot;6&quot;, 12}, {&quot;97&quot;, 194}, {&quot;9&quot;, 18}, {&quot;99&quot;, 198}, {&quot;10&quot;, 20}, {&quot;100&quot;, 200} ] Got batch: [ {&quot;29&quot;, 58}, {&quot;32&quot;, 64}, ... ]","ref":"rabbitmq.html#run-the-broadway-pipeline","title":"RabbitMQ - Run the Broadway pipeline","type":"extras"},{"doc":"Some of the configuration options available for Broadway come already with a &quot;reasonable&quot; default value. However, those values might not suit your requirements. Depending on the number of messages you get, how much processing they need and how much IO work is going to take place, you might need completely different values to optimize the flow of your pipeline. The stages option available for every set of producers, processors and batchers, among with batch_size and batch_timeout can give you a great deal of flexibility. The stages option controls the concurrency level in each layer of the pipeline. Another important option to take into account is the :prefetch_count. Note that unlike the RabittMQ client that has a default :prefetch_count = 0, which disables back-pressure, BroadwayRabbitMQ overwrite the default value to 50 enabling the back-pressure mechanism. You can see more details in the &quot;Back-pressure and :prefetch_count&quot; section of the BroadwayRabbitMQ documentation. In order to get a good set of configurations for your pipeline, it&#39;s important to respect the limitations of the servers you&#39;re running, as well as the limitations of the services you&#39;re providing/consuming data to/from.","ref":"rabbitmq.html#tuning-the-configuration","title":"RabbitMQ - Tuning the configuration","type":"extras"},{"doc":"Custom Producers If you want to use Broadway but there is no existing Broadway producer for the technology of your choice, you can integrate any existing GenStage producer into the pipeline with relative ease.","ref":"custom-producers.html","title":"Custom Producers","type":"extras"},{"doc":"In general, producers must generate %Broadway.Message{} structs in order to be processed by Broadway. In case you need to use an existing GenStage producer and you don&#39;t want to change its original implementation, you&#39;ll have to set the producer&#39;s :transformer option to translate the generated events into Broadway messages. In the following example the producer is a regular GenStage, i.e., it produces plain events that cannot be processed by Broadway directly: defmodule Counter do use GenStage def start_link(number) do GenStage.start_link(Counter, number) end def init(counter) do {:producer, counter} end def handle_demand(demand, counter) when demand &gt; 0 do events = Enum.to_list(counter..counter+demand-1) {:noreply, events, counter + demand} end end By using a transformer, you can tell Broadway to transform all events generated by the producer into proper Broadway messages: defmodule MyBroadway do use Broadway alias Broadway.Message def start_link(_opts) do Broadway.start_link(__MODULE__, name: __MODULE__, producer: [ module: {Counter, 1}, transformer: {__MODULE__, :transform, []} ], processors: [ default: [concurrency: 10] ], batchers: [ default: [concurrency: 2, batch_size: 5], ] ) end ...callbacks... def transform(event, _opts) do %Message{ data: event, acknowledger: {__MODULE__, :ack_id, :ack_data} } end def ack(:ack_id, successful, failed) do # Write ack code here end end Notice that you need to pass two options to the producer: :module - a tuple representing the GenStage producer as {mod, arg}. Where mod is module that implements the GenStage behaviour and arg the argument that will be given to the init callback of the GenStage. It is very important to note that Broadway will not call the child_spec/1 or start_link/1 function of the producer. That&#39;s because Broadway wraps the producer to augment it with extra features. :transformer - a module-function-args tuple that will be invoked inside the producer, for every producer message, that should create a Broadway.Message struct with the data and acknowledger fields. See the Broadway.Acknowledger module for more information on defining and setting up acknowledgements.","ref":"custom-producers.html#example","title":"Custom Producers - Example","type":"extras"},{"doc":"Architecture Broadway&#39;s architecture is built on top of GenStage. That means we structure our processing units as independent stages that are responsible for one individual task in the pipeline. By implementing the Broadway behaviour, we define a GenServer process that wraps a Supervisor to manage and own our pipeline.","ref":"architecture.html","title":"Architecture","type":"extras"},{"doc":"[producers] &lt;- pulls data from SQS, RabbitMQ, etc. | | (demand dispatcher) | handle_message/3 runs here -&gt; [processors] / \\ / \\ (partition dispatcher) / \\ [batcher] [batcher] &lt;- one for each batcher key | | | | (demand dispatcher) | | handle_batch/4 runs here -&gt; [batch processor][batch processor]","ref":"architecture.html#the-pipeline-model","title":"Architecture - The pipeline model","type":"extras"},{"doc":"Broadway.Producer - A wrapper around the actual producer defined by the user. It serves as the source of the pipeline. Broadway.Processor - This is where messages are processed, e.g. do calculations, convert data into a custom json format etc. Here is where the code from handle_message/3 runs. Broadway.Batcher - Creates batches of messages based on the batcher&#39;s key. One batcher for each key will be created. Broadway.Consumer - This is where the code from handle_batch/4 runs.","ref":"architecture.html#internal-stages","title":"Architecture - Internal stages","type":"extras"},{"doc":"Broadway was designed to always go back to a working state in case of failures thanks to the use of supervisors. Our supervision tree is designed as follows: [Broadway GenServer] | | | [Broadway Pipeline Supervisor] / / (:rest_for_one) \\ \\ / | | \\ / | | \\ / | | \\ / | | \\ / | | \\ [ProducerSupervisor] [ProcessorSupervisor] [BatcherPartitionSupervisor] [Terminator] (:one_for_one) (:one_for_all) (:one_for_one) / \\ / \\ / \\ / \\ / \\ / \\ / \\ / \\ / \\ / \\ / \\ / \\ [Producer_1] ... [Processor_1] ... [BatcherConsumerSuperv_1] ... (:rest_for_one) / \\ / \\ / \\ [Batcher] [ConsumerSupervisor] (:one_for_all) / \\ / \\ / \\ [Consumer_1] ... Both ProcessorSupervisor and ConsumerSupervisor are set with max_restarts to 0. The idea is that if any process fails, we want to restart the rest of the tree. Since Broadway callbacks are stateless, we can handle errors and provide reports without crashing processes. This means that the supervision tree will only shutdown in case of unforeseen errors in Broadway&#39;s implementation. The only exception are the producers, which contain external code and are expected to fail. If a producer crashes, it will be restarted by its supervisor without cascading failures until its max restarts is reached. Broadway automatically handles those failures by making processors automatically resubscribe to producers in case of crashes.","ref":"architecture.html#the-supervision-tree","title":"Architecture - The supervision tree","type":"extras"},{"doc":"The cascading failures aspect also provides safe semantics for graceful shutdown. We know that either all processes are running OR they are all being shutdown. Therefore, to gracefully shutdown the supervision tree, a terminator process is activated, which starts the following steps: It notifies the first layer of processors that they should not resubscribe to producers once they exit It tells all producers to no longer accept demand, flush all current events, and then shutdown It then monitors and waits for a confirmation message from batch processors. At this point, the terminator is effectively blocking the supervisor until all events have been processed This triggers a cascade effect where processors notice all of its producers have been cancelled, causing them to flush their own events and cancels the stages downstream, and so on and so on. This happens until batch processors notice all of their producers have been cancelled, effectively notifying the terminator to shutdown, allowing the outer most supervisor to go on and fully terminate all stages, which at this point have flushed all events.","ref":"architecture.html#graceful-shutdowns","title":"Architecture - Graceful shutdowns","type":"extras"}]